--[[

--]]
--
-- An FRSKY S.Port <passthrough protocol> based Telemetry script for the Horus X10 and X12 radios
--
-- Copyright (C) 2018-2019. Alessandro Apostoli
-- https://github.com/yaapu
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY, without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, see <http://www.gnu.org/licenses>.
--

---------------------
-- MAIN CONFIG
-- 480x272 LCD_W x LCD_H
---------------------

---------------------
-- VERSION
---------------------
-- load and compile of lua files
--#define LOADSCRIPT
-- uncomment to force compile of all chunks, comment for release
--#define COMPILE
-- fix for issue OpenTX 2.2.1 on X10/X10S - https://github.com/opentx/opentx/issues/5764

---------------------
-- MAVLITE CONFIG
---------------------

---------------------
-- DEV FEATURE CONFIG
---------------------
-- enable sport debug
-- enable mavlite debug
-- enable events debug
--#define DEBUGEVT
-- cache tuning pages
-- cache params pages
-- enable full telemetry debug
--#define TELEMETRY_DEBUG
-- enable full telemetry decoding
--#define FULL_TELEMETRY
-- enable memory debuging 
-- enable dev code
--#define DEV
-- use radio channels imputs to generate fake telemetry data
--#define TESTMODE

---------------------
-- DEBUG REFRESH RATES
---------------------
-- calc and show hud refresh rate
-- calc and show telemetry process rate



--------------------------------------------------------------------------------
-- MENU VALUE,COMBO
--------------------------------------------------------------------------------

-----------------------
-- LIBRARY LOADING
-----------------------

--[[
  status of pending mavlite messages
]]





-- enable sport debug
-- enable mavlite debug



local frameTypes = {}
local frameType = nil

--[[
	MAV_TYPE_GENERIC=0,               /* Generic micro air vehicle. | */
	MAV_TYPE_FIXED_WING=1,            /* Fixed wing aircraft. | */
	MAV_TYPE_QUADROTOR=2,             /* Quadrotor | */
	MAV_TYPE_COAXIAL=3,               /* Coaxial helicopter | */
	MAV_TYPE_HELICOPTER=4,            /* Normal helicopter with tail rotor. | */
	MAV_TYPE_ANTENNA_TRACKER=5,       /* Ground installation | */
	MAV_TYPE_GCS=6,                   /* Operator control unit / ground control station | */
	MAV_TYPE_AIRSHIP=7,               /* Airship, controlled | */
	MAV_TYPE_FREE_BALLOON=8,          /* Free balloon, uncontrolled | */
	MAV_TYPE_ROCKET=9,                /* Rocket | */
	MAV_TYPE_GROUND_ROVER=10,         /* Ground rover | */
	MAV_TYPE_SURFACE_BOAT=11,         /* Surface vessel, boat, ship | */
	MAV_TYPE_SUBMARINE=12,            /* Submarine | */
  MAV_TYPE_HEXAROTOR=13,            /* Hexarotor | */
	MAV_TYPE_OCTOROTOR=14,            /* Octorotor | */
	MAV_TYPE_TRICOPTER=15,            /* Tricopter | */
	MAV_TYPE_FLAPPING_WING=16,        /* Flapping wing | */
	MAV_TYPE_KITE=17,                 /* Kite | */
	MAV_TYPE_ONBOARD_CONTROLLER=18,   /* Onboard companion controller | */
	MAV_TYPE_VTOL_DUOROTOR=19,        /* Two-rotor VTOL using control surfaces in vertical operation in addition. Tailsitter. | */
	MAV_TYPE_VTOL_QUADROTOR=20,       /* Quad-rotor VTOL using a V-shaped quad config in vertical operation. Tailsitter. | */
	MAV_TYPE_VTOL_TILTROTOR=21,       /* Tiltrotor VTOL | */
	MAV_TYPE_VTOL_RESERVED2=22,       /* VTOL reserved 2 | */
	MAV_TYPE_VTOL_RESERVED3=23,       /* VTOL reserved 3 | */
	MAV_TYPE_VTOL_RESERVED4=24,       /* VTOL reserved 4 | */
	MAV_TYPE_VTOL_RESERVED5=25,       /* VTOL reserved 5 | */
	MAV_TYPE_GIMBAL=26,               /* Onboard gimbal | */
	MAV_TYPE_ADSB=27,                 /* Onboard ADSB peripheral | */
	MAV_TYPE_PARAFOIL=28,             /* Steerable, nonrigid airfoil | */
	MAV_TYPE_DODECAROTOR=29,          /* Dodecarotor | */
]]
-- copter
frameTypes[0]   = "c"
frameTypes[2]   = "c"
frameTypes[3]   = "c"
frameTypes[4]   = "h"
frameTypes[13]  = "c"
frameTypes[14]  = "c"
frameTypes[15]  = "c"
frameTypes[29]  = "c"
-- plane
frameTypes[1]   = "p"
frameTypes[16]  = "p"
frameTypes[19]  = "q"
frameTypes[20]  = "q"
frameTypes[21]  = "q"
frameTypes[22]  = "q"
frameTypes[23]  = "q"
frameTypes[24]  = "q"
frameTypes[25]  = "q"
frameTypes[28]  = "p"
-- rover
frameTypes[10]  = "r"
-- boat
frameTypes[11]  = "b"

------------------------------
-- TELEMETRY DATA
------------------------------
local telemetry = {}
-- PARAMS
telemetry.frameType = -1

--[[
0	MAV_SEVERITY_EMERGENCY	System is unusable. This is a "panic" condition.
1	MAV_SEVERITY_ALERT	Action should be taken immediately. Indicates error in non-critical systems.
2	MAV_SEVERITY_CRITICAL	Action must be taken immediately. Indicates failure in a primary system.
3	MAV_SEVERITY_ERROR	Indicates an error in secondary/redundant systems.
4	MAV_SEVERITY_WARNING	Indicates about a possible future error if this is not resolved within a given timeframe. Example would be a low battery warning.
5	MAV_SEVERITY_NOTICE	An unusual event has occured, though not an error condition. This should be investigated for the root cause.
6	MAV_SEVERITY_INFO	Normal operational messages. Useful for logging. No action is required for these messages.
7	MAV_SEVERITY_DEBUG	Useful non-operational messages that can assist in debugging. These should not occur during normal operation.
--]]
local mavSeverity = {}

mavSeverity[0]="EMR"
mavSeverity[1]="ALR"
mavSeverity[2]="CRT"
mavSeverity[3]="ERR"
mavSeverity[4]="WRN"
mavSeverity[5]="NOT"
mavSeverity[6]="INF"
mavSeverity[7]="DBG"

--------------------------------
-- STATUS DATA
--------------------------------
local status = {}
-- MESSAGES
status.messages = {}
status.msgBuffer = ""
status.lastMsgValue = 0
status.lastMsgTime = 0
status.lastMessage = nil
status.lastMessageSeverity = 0
status.lastMessageCount = 1
status.messageCount = 0
-- LINK STATUS
status.noTelemetryData = 1
status.hideNoTelemetry = false


local soundFileBasePath = "/SOUNDS/yaapu0"

----------------------
--- COLORS
----------------------

--#define COLOR_LABEL 0x7BCF
--#define COLOR_BG 0x0169




local hudcounter = 0
local hudrate = 0
local hudstart = 0

local maxmem = 0

--------------------------------------------------------------------------------
-- CONFIGURATION MENU
--------------------------------------------------------------------------------
local conf = {
  language = "en"
}

local utils = {}
local mavLib = {}

local basePath = "/SCRIPTS/YAAPU/"
local libBasePath = basePath.."LIB/"

local bitmaps = {}
local blinktime = getTime()
local blinkon = false

local menu  = {
  selectedItem = 1,
  editSelected = false,
  offset = 0,
  wrapOffset = 0, -- changes according to enabled/disabled features and panels
  page = 0,
}

local sportPacket = {
  sensor_id = nil,
  frame_id = nil,
  data_id = nil,
  value = nil
}

local mavlite_message= {
  msgid = -1,
  len = 0,
  payload = {},
  checksum = 0
}

local mavlite_status = {
  parse_state = 0,
  current_rx_seq = 0,
  payload_next_byte = 0
}

local idx = 1 -- index of the current item being processed by processItems()
local page = 1
local subPage = 0

utils.doLibrary = function(filename)
  local f = assert(loadScript(libBasePath..filename..".lua"))
  collectgarbage()
  collectgarbage()
  return f()
end

-----------------------------
-- clears the loaded table 
-- and recovers memory
-----------------------------
utils.clearTable = function(t)
  if type(t)=="table" then
    for i,v in pairs(t) do
      if type(v) == "table" then
        utils.clearTable(v)
      end
      t[i] = nil
    end
  end
  t = nil
  collectgarbage()
  collectgarbage()
  maxmem = 0
end  

local function getItemByName(items,name)
  for idx=1,#items
  do
    if items[idx][1] == name then
      return items[idx]
    end
  end
  return nil
end

utils.drawBottomBar = function()
  lcd.setColor(CUSTOM_COLOR,0x0000)
  -- black bar
  lcd.drawFilledRectangle(0,LCD_H-20+3, LCD_W, 17, CUSTOM_COLOR)
  -- message text
  local msg = status.messages[(status.messageCount + #status.messages) % (#status.messages+1)][1]
  
  if status.messages[(status.messageCount + #status.messages) % (#status.messages+1)][2] < 4 then
    lcd.setColor(CUSTOM_COLOR,lcd.RGB(255,70,0))  
  elseif status.messages[(status.messageCount + #status.messages) % (#status.messages+1)][2] == 4 then
      lcd.setColor(CUSTOM_COLOR,lcd.RGB(255,255,0))  
  else
    lcd.setColor(CUSTOM_COLOR,0xFFFF)
  end
  lcd.drawText(1, LCD_H - 20, msg, CUSTOM_COLOR)
end

local function telemetryEnabled()
  if getRSSI() == 0 then
    status.noTelemetryData = 1
  end
  return status.noTelemetryData == 0
end

local function drawWarning(text)
  lcd.setColor(CUSTOM_COLOR,0xFFFF)
  lcd.drawFilledRectangle(48,74, 384, 84, CUSTOM_COLOR)
  lcd.setColor(CUSTOM_COLOR,0xF800)
  lcd.drawFilledRectangle(50,76, 380, 80, CUSTOM_COLOR)
  lcd.setColor(CUSTOM_COLOR,0xFFFF)
  lcd.drawText(65, 80, text, DBLSIZE+CUSTOM_COLOR)
  lcd.drawText(130, 130, "Yaapu Ardupilot LuaGCS 0.3-dev", SMLSIZE+CUSTOM_COLOR)
end

local function drawBars(items)
  lcd.setColor(CUSTOM_COLOR,0x0000)
  lcd.drawFilledRectangle(0,0, LCD_W, 20, CUSTOM_COLOR)
  lcd.drawRectangle(0, 0, LCD_W, 20, CUSTOM_COLOR)
  local itemIdx = string.format("%d/%d",menu.selectedItem,#items)
  lcd.setColor(CUSTOM_COLOR,0xFFFF)
  lcd.drawText(LCD_W,LCD_H-20+1,itemIdx,CUSTOM_COLOR+RIGHT)
end

local function drawMessageScreen()
  for i=0,#status.messages do
    if  status.messages[(status.messageCount + i) % (#status.messages+1)][2] == 4 then
      lcd.setColor(CUSTOM_COLOR,lcd.RGB(255,255,0))
    elseif status.messages[(status.messageCount + i) % (#status.messages+1)][2] < 4 then
      lcd.setColor(CUSTOM_COLOR,lcd.RGB(255,70,0))  
    else
      lcd.setColor(CUSTOM_COLOR,0xFFFF)
    end
    lcd.drawText(0,2+13*i, status.messages[(status.messageCount + i) % (#status.messages+1)][1],SMLSIZE+CUSTOM_COLOR)
  end
end

local function processMavliteMessage(msg)
  if msg.msgid == 23 then -- PARAM_SET
    local param_name = mavLib.msg_get_name(msg,4)
    local param_value = mavLib.msg_get_value(msg,0)
    utils.pushMessage(7,string.format("RX(MSG): id=%d, %s : %f",msg.msgid,param_name,param_value), true)
  elseif msg.msgid == 22 then -- PARAM_VALUE
    local param_name = mavLib.msg_param_value_get_name(msg)
    local param_value = mavLib.msg_param_value_get_value(msg)
    utils.pushMessage(7,string.format("RX(MSG): ID=%d, %s : %f",msg.msgid,param_name,param_value))
  elseif msg.msgid == 76 then -- COMMAND_LONG
    local cmd_id = mavLib.msg_get_uint16(msg,0)
    local options = mavLib.msg_get_uint8(msg,2)
    local param_count = mavLib.bit8_unpack(options,3,0)
    utils.pushMessage(7, string.format("RX(MSG) id:%d, len:%d, cmd_id:%d, param_count:%d",msg.msgid, msg.len, cmd_id, param_count), true)
  elseif msg.msgid == 77 then -- CMD_ACK
    local cmd_id = mavLib.msg_get_uint16(msg,0)
    local cmd_result = mavLib.msg_get_uint8(msg,2)
    utils.pushMessage(7, string.format("RX(MSG) id:%d, len:%d, cmd_id:%d, cmd_result:%d",msg.msgid, msg.len, cmd_id,cmd_result), true)
  end
end

local function formatMessage(severity,msg)
  local clippedMsg = msg
  
  if #msg > 50 then
    clippedMsg = string.sub(msg,1,50)
    msg = nil
    collectgarbage()
    collectgarbage()
  end
  
  if status.lastMessageCount > 1 then
    return string.format("%02d:%s (x%d) %s", status.messageCount, mavSeverity[severity], status.lastMessageCount, clippedMsg)
  else
    return string.format("%02d:%s %s", status.messageCount, mavSeverity[severity], clippedMsg)
  end
end

utils.getBitmap = function(name)
  if bitmaps[name] == nil then
    bitmaps[name] = Bitmap.open("/SCRIPTS/YAAPU/IMAGES/"..name..".png")
  end
  return bitmaps[name],Bitmap.getSize(bitmaps[name])
end

utils.unloadBitmap = function(name)
  if bitmaps[name] ~= nil then
    bitmaps[name] = nil
    -- force call to luaDestroyBitmap()
    collectgarbage()
    collectgarbage()
  end
end

utils.drawBlinkBitmap = function(bitmap,x,y)
  if blinkon == true then
      lcd.drawBitmap(utils.getBitmap(bitmap),x,y)
  end
end

utils.playSound = function(soundFile,skipHaptic)
  playFile(soundFileBasePath .."/"..conf.language.."/".. soundFile..".wav")
end

utils.pushMessage = function(severity, msg, silent)
  if severity < 5 and silent == nil then
    utils.playSound("../err",true)
  else
    utils.playSound("../inf",true)
  end
  
  if msg == status.lastMessage then
    status.lastMessageCount = status.lastMessageCount + 1
  else  
    status.lastMessageCount = 1
    status.messageCount = status.messageCount + 1
  end
  if status.messages[(status.messageCount-1) % 20] == nil then
    status.messages[(status.messageCount-1) % 20] = {}
  end
  status.messages[(status.messageCount-1) % 20][1] = formatMessage(severity,msg)
  status.messages[(status.messageCount-1) % 20][2] = severity
  
  status.lastMessage = msg
  status.lastMessageSeverity = severity
  -- Collect Garbage
  collectgarbage()
  collectgarbage()
end

local function processTelemetry(sp)
  if sp.data_id == 0x5000 then -- MESSAGES
    if sp.value ~= status.lastMsgValue then
      status.lastMsgValue = sp.value
      local c
      local msgEnd = false
      for i=3,0,-1
      do
        c = bit32.extract(sp.value,i*8,7)
        if c ~= 0 then
          status.msgBuffer = status.msgBuffer .. string.char(c)
          collectgarbage()
          collectgarbage()
        else
          msgEnd = true;
          break;
        end
      end
      if msgEnd then
        local severity = (bit32.extract(sp.value,7,1) * 1) + (bit32.extract(sp.value,15,1) * 2) + (bit32.extract(sp.value,23,1) * 4)
        utils.pushMessage( severity, status.msgBuffer)
        status.msgBuffer = nil
        -- recover memory
        collectgarbage()
        collectgarbage()
        status.msgBuffer = ""
      end
    end
  end
end

local function processSportData(sportPacket)
  mavLib.process_sport_rx_data(mavlite_message, mavlite_status, processMavliteMessage, sportPacket)
end

local function createMsgParamRequestRead(paramName)
    
    local msg = {
      msgid = 20,
      len = 0,
      payload = {},
      checksum = 0
    }

    mavLib.msg_param_request_read_set_name(msg,paramName)
    
    collectgarbage()
    collectgarbage()
    
    return msg
end

local function createMsgParamSet(paramName, paramValue)
    
    local msg = {
      msgid = 23,
      len = 0,
      payload = {},
      checksum = 0
    }

    mavLib.msg_param_set_set_value(msg, paramValue)
    mavLib.msg_param_set_set_name(msg, paramName)
    
    collectgarbage()
    collectgarbage()
    
    return msg
end

local sendMavliteTimer = getTime()

local refreshTimer = 0

local function background()
  for i=1,20
  do
    local sensor_id,frame_id,data_id,value = sportTelemetryPop()
    
    if sensor_id  ~= nil then
      
      sportPacket.sensor_id = sensor_id
      sportPacket.frame_id = frame_id
      sportPacket.data_id = data_id
      sportPacket.value = value
      
      if sportPacket.frame_id == 0x10 then
        status.noTelemetryData = 0
        -- no telemetry dialog only shown once
        status.hideNoTelemetry = true
        
        processTelemetry(sportPacket)
      elseif sportPacket.frame_id == 0x32 then
        utils.pushMessage(7,string.format("RX %02X:%02X:%04X:%08X",sportPacket.sensor_id, sportPacket.frame_id, sportPacket.data_id, sportPacket.value))
        processSportData(sportPacket)
      end
    end  
  end
  
  
  if getTime() - sendMavliteTimer > 500 then
    
    local msg = {
      msgid = 77, -- command long
      len = 0,
      payload = {},
      checksum = 0
    }
    
    local options = 0
    -- set param count to 3
    options = mavLib.bit8_pack(options,3,3,0)
    -- command 241: calibrate
    mavLib.msg_set_uint16(msg,241,0)
    -- declare 3 parameters
    mavLib.msg_set_uint8(msg,241,2)
    -- param 1
    mavLib.msg_set_float(msg,0,3)
    -- param 2
    mavLib.msg_set_float(msg,0,3+4)
    -- param 3
    mavLib.msg_set_float(msg,0,3+8)
    
    local success = mavLib.msg_send(msg,utils)
    
    utils.pushMessage((success and 7 or 4),"TX(MSG): COMMAND_LONG(241) "..(success and "OK" or "KO"))
    
    sendMavliteTimer = getTime()
  end
  
  -- send pending packets
  mavLib.process_sport_tx_queue(utils)
  
  collectgarbage()
  collectgarbage()
end

local function getModelFilename()
  local info = model.getInfo()
  return "/SCRIPTS/YAAPU/CFG/" .. string.lower(string.gsub(info.name, "[%c%p%s%z]", ""))
end

local showMessageScreen = false
--------------------------
-- RUN
--------------------------
local function run(event)
  ------------------------
  -- CALC HUD REFRESH RATE
  ------------------------
  local hudnow = getTime()
  
  if hudcounter == 0 then
    hudstart = hudnow
  else
    hudrate = hudrate*0.8 + 100*(hudcounter/(hudnow - hudstart + 1))*0.2
  end
  
  hudcounter=hudcounter+1
  
  if hudnow - hudstart + 1 > 1000 then
    hudcounter = 0
  end
  
  background()
  
  if showMessageScreen == true then
    lcd.setColor(CUSTOM_COLOR, 0x0000)
  else
    lcd.setColor(CUSTOM_COLOR, 0x0AB1)
  end
  
  lcd.clear(CUSTOM_COLOR)
  
  drawMessageScreen()
  
  utils.drawBottomBar()
  
  -- no telemetry/minmax outer box
  if telemetryEnabled() == false then
    utils.drawBlinkBitmap("warn",0,0)  
  end

  lcd.setColor(CUSTOM_COLOR,0xFE60)
  local hudrateTxt = string.format("%.1ffps",hudrate)
  lcd.drawText(480,3,hudrateTxt,SMLSIZE+CUSTOM_COLOR+RIGHT)
  
  lcd.setColor(CUSTOM_COLOR,lcd.RGB(255,0,0))
  maxmem = math.max(maxmem,collectgarbage("count")*1024)
  -- test with absolute coordinates
  lcd.drawNumber(450,LCD_H-14,maxmem,SMLSIZE+MENU_TITLE_COLOR+RIGHT)

  return 0
end

local function init()
  -- load mavlite library
  mavLib = utils.doLibrary("mavlite")  
  -- ok done
  utils.pushMessage(7,"Yaapu Ardupilot LuaGCS 0.3-dev")
  collectgarbage()
  collectgarbage()
end

--------------------------------------------------------------------------------
-- SCRIPT END
--------------------------------------------------------------------------------
return {run=run, init=init}
